#include <stdio.h>
#include <string.h>
#include <isql.h>
#include <isqlext.h>

static HENV env;        /* SQL environment */
static HDBC connect;    /* SQL connection */
static HSTMT statement; /* SQL statement */
static SWORD APILevel;  /* API conformance level */
static SWORD SQLLevel;  /* ODBC conformance level */

/*
 * String der Form "DSN=xxx;UID=xxx;PWD=xxx" in seine
 * Bestandteile zerlegen
 */
void odbc_split_dsn(char *dsnin,
		    char **dsnout,
		    char **uid,
		    char **pwd)
{
#define SEP ';'
#define UID "UID"
#define PWD "PWD"
#define DSN "DSN"
  register int i;
  
  char *tok[6];
  char *p,*q;
  char *copy = strdup(dsnin);
  *pwd = *dsnout = *uid = NULL;
  tok[0] = strtok(copy, ";");
  tok[1] = strtok(NULL,";");
  tok[2] = strtok(NULL, ";");
  tok[3] = strtok(NULL, ";");
  tok[4] = strtok(NULL, ";");
  tok[5] = NULL;
  i = 0;
  while (tok[i]) {
    q = strtok(tok[i],"=");
    if (strcmp(DSN,q) == 0) {
      *dsnout = strdup(q+strlen(DSN) + 1);
    } else if (strcmp(UID,q) == 0) {
      *uid = strdup(q + strlen(DSN) + 1);
    } else if (strcmp(PWD,q) == 0) {
      *pwd = strdup(q + strlen(DSN) + 1);
    } 
    i++;
  }
  free(copy);
}

/*
 * Alle Tabellen der aktuellen Datenbank 
 * per ODBC erfragen und zurckliefern
 */
static char **get_tables(int *count)
{
  char tableName[129];
  int  nameLength, SQLret;
  int  maxel = 10;

  char **retval = (char **)malloc(sizeof(char *) 
				  * maxel);
  
  *count = 0;
  
  SQLret = SQLTables(statement,"%",SQL_NTS,
		     "%", SQL_NTS,
		     "%", SQL_NTS,
		     "TABLE",SQL_NTS);
  if (SQL_SUCCESS != SQLret)
    odbc_print_error(__LINE__,__FILE__,NULL, 
		     NULL, statement);
  
  SQLret = SQLBindCol(statement, 3, SQL_C_CHAR, 
		      (SQLPOINTER) tableName,
		      128, &nameLength);
  if (SQL_SUCCESS != SQLret)
    odbc_print_error(__LINE__,__FILE__,NULL, 
		     NULL, statement);
  
  while (SQL_SUCCESS    == (SQLret = SQLFetch(statement)) ||
	 SQL_SUCCESS_WITH_INFO == SQLret) {
    if (*count >= maxel) {
      retval = (char **)realloc(retval, maxel * 2);
      maxel *= 2;
    }

    retval[*count] = (char *)calloc(sizeof(char *),
				    nameLength + 1);
    strncpy(retval[*count],tableName,nameLength);
    (*count)++;
    
  }
  if (*count == 0) {
    free(retval);
    retval = NULL;
  } else {
    retval = (char **)realloc(retval, *count);
  }
  return retval;
}

/*
 * Verbindung zur Datenbank herstellen
 */
int odbc_open(char *dsn)
{
  register int i;
  int SQLret, outputLength, tableCount;
  SWORD ignore;
  char *uid, *pwd, *newdsn, *outputString;
  char **tables;
  BOOL driverConnect;
  BOOL done;

  SQLAllocEnv(&env);

  SQLret = SQLAllocConnect(env, &connect);

  if (SQL_SUCCESS != SQLret)
    odbc_print_error(__LINE__,__FILE__,NULL ,
		     connect, SQL_NULL_HSTMT);

  /*
   * Unterstuetzt der Treibermanager 
   * SQLDriverConnect()?
   */
  SQLGetFunctions(connect,SQL_API_SQLDRIVERCONNECT,
		  &driverConnect);
  done = 0;
  if (driverConnect) {
    outputLength = 255;
    outputString = (char *)malloc(outputLength + 1);
    printf("Using SQLDriverConnect <%s>\n",dsn);
    SQLret = SQLDriverConnect(connect,(HWND)0,
			      dsn,SQL_NTS,
			      outputString,
			      outputLength,
			      &outputLength,
			      SQL_DRIVER_NOPROMPT);
    if (SQL_SUCCESS ==SQLret) {
      strcpy(dsn, outputString);
      done = 1;
    } else {
      SQLDisconnect(env,connect);
      SQLAllocConnect(env, &connect);
    }
  } 
  /*
   * Es gibt ODBC-Treiber, die SQLDriverConnect _nicht_
   * unterstuetzen, obwohl der Treibermanager es kann.
   * In diesem Fall ist "done" immer noch falsch
   */
  if (!done) {
    odbc_split_dsn(dsn,&newdsn,&uid, &pwd); 
    printf("Using SQLConnect DSN <%s> UID <%s>\n",
	   newdsn, uid);
    SQLret = SQLConnect(connect,
			newdsn, SQL_NTS,
			uid ? uid : NULL, SQL_NTS,
			pwd ? pwd : NULL, SQL_NTS);
  }
  if (SQL_SUCCESS != SQLret) {
    odbc_print_error(__LINE__,__FILE__,NULL,
		     connect, SQL_NULL_HSTMT);
    return 0;
  }
  /*
   * Statement-Handle besorgen und 
   * die ODBC/SQL-Konformitt erfragen
   */
  SQLAllocStmt(connect,&statement);
  SQLGetInfo(connect,
	     SQL_ODBC_API_CONFORMANCE,
	     &APILevel,
	     sizeof (APILevel),
	     &ignore);
  SQLGetInfo(connect,
	     SQL_ODBC_SQL_CONFORMANCE,
	     &SQLLevel,
	     sizeof (SQLLevel),
	     &ignore);
  printf("APILevel: %d, ODBC/SQLLevel: %d\n"
	 APILevel, SQLLevel);
  /*
   * Tabellen erfragen und ausgeben
   */
  tables = get_tables(&tableCount);
  for (i = 0; i < tableCount; i++)
    printf("Table %d: <%s>\n", i+1, tables[i]);
  return 1;
}


/*
 * Fehlermeldungen ausgeben
 */
void odbc_print_error(int line,char *file, 
     HENV henv, HDBC hdbc, HSTMT statement)
{
  char state[256],message[SQL_MAX_MESSAGE_LENGTH - 1];
  SDWORD native;
  SWORD  messageLength;

  SQLError(henv, hdbc, statement,
	   state,&native,message,sizeof(message),
	   &messageLength);
	   
  fprintf(stderr,"SQL-Error in %s, line %d\t%s\n\t\"%s\"\n",
	  file,line, state, message);
  if (native != 0)
    fprintf(stderr,"\t[Native error code %d]\n", native);
}
			      
main() {
  char datasource[2048];
  printf("DSN: \n");
  scanf("%s",datasource);
  odbc_open(datasource);
}
